'''
    @author neucrack
    @license LGPL-3.0
    @update 2022-03-15 create
'''
from ast import arg
from PyQt5.QtCore import QObject, Qt, QRect
from PyQt5.QtWidgets import (QApplication, QWidget,QPushButton,QMessageBox,QDesktopWidget,QMainWindow,
                             QVBoxLayout,QHBoxLayout,QGridLayout,QTextEdit,QLabel,QRadioButton,QCheckBox,
                             QLineEdit,QGroupBox,QSplitter,QFileDialog, QScrollArea, QScrollBar)
from PyQt5.QtGui import (QColor, QPixmap, QFontMetrics, QFont, QBrush, QPen, QPainter,
                        QKeyEvent, QWheelEvent)

try:
    from .base import Plugin_Base
    from Combobox import ComboBox
    from i18n import _
    import utils, parameters
    from conn.base import ConnectionStatus
except ImportError:
    from COMTool.plugins.base import Plugin_Base
    from COMTool import utils, parameters
    from COMTool.i18n import _
    from COMTool.Combobox import ComboBox
    from COMTool.conn.base import  ConnectionStatus

import pyte
from wcwidth import wcswidth
import threading
import time
import pyperclip

class Terminal_Backend:
    def __init__(self) -> None:
        self.screen = pyte.HistoryScreen(60, 20, history=9999, ratio = 0.05)
        self.screen.set_mode(pyte.modes.LNM)
        self.stream = pyte.ByteStream(self.screen)

    def feed(self, data:bytes):
        self.stream.feed(data)

    def resize(self, width, height):
        self.screen.resize(columns=width, lines=height)
        self.screen.history = self.screen.history._replace(ratio = 1.0/height)

class Terminal_Frontend(QWidget):
    def __init__(self, send_func, resizeConnFunc, isConnectedFunc) -> None:
        super().__init__()
        self.send = send_func
        self.resizeConn = resizeConnFunc
        self.isConnected = isConnectedFunc
        self.theme = {
            "bg": QColor("#212121"),
            "color": QColor(255, 255, 255),
            "cursor": QColor(76, 175, 80),
            "colors": {
                'black': QColor("#212121"),
                'red': QColor("#f44336"),
                'green': QColor("#4caf50"),
                'blue': QColor("#2196f3"),
                'cyan': QColor("#26c6da"),
                'brown': QColor("#a1887f"),
                'yellow': QColor("#ffeb3b"),
                'magenta': QColor("#e85aad"),
                'white': QColor("#cacaca"),
                "selected": QColor(255, 255, 255, 80),
            }
        }
        self.keymap = {
            # for term type string equal to linux
            Qt.Key_Backspace: b'\x7F', # use DEL, not 0x08(^H, Backspace)
            Qt.Key_Escape:    b'\x1B',
            Qt.Key_AsciiTilde: b'\x7E',
            Qt.Key_ScrollLock: b'\x1A',
            Qt.Key_Up:       b'\x1b[A',
            Qt.Key_Down:     b'\x1b[B',
            Qt.Key_Left:     b'\x1b[D',
            Qt.Key_Right:    b'\x1b[C',
            Qt.Key_PageUp:   b"\x1b[5~",
            Qt.Key_PageDown: b"\x1b[6~",
            Qt.Key_Home:     b'\x1b[1~',
            Qt.Key_End:      b'\x1b[4~',
            Qt.Key_Insert:   b"\x1b[2~",
            Qt.Key_Delete:   b"\x1b[3~",
            Qt.Key_F1:       b"\x1b[11~",
            Qt.Key_F2:       b"\x1b[12~",
            Qt.Key_F3:       b"\x1b[13~",
            Qt.Key_F4:       b"\x1b[14~",
            Qt.Key_F5:       b"\x1b[15~",
            Qt.Key_F6:       b"\x1b[17~",
            Qt.Key_F7:       b"\x1b[18~",
            Qt.Key_F8:       b"\x1b[19~",
            Qt.Key_F9:       b"\x1b[20~",
            Qt.Key_F10:      b"\x1b[21~",
            Qt.Key_F11:      b"\x1b[23~",
            Qt.Key_F12:      b"\x1b[24~",
            Qt.Key_F13:      b"\x1b[25~",
        }
        self.setAttribute(Qt.WA_InputMethodEnabled)
        self.resize(580, 380)
        self.setMinimumWidth(40)
        self.setMinimumHeight(40)
        self.backend = Terminal_Backend()
        self.setCursor(Qt.IBeamCursor)
        self.setFocusPolicy(Qt.StrongFocus)
        self.font_name = 'Consolas,Bitstream Vera Sans Mono,Courier New,monospace, Microsoft Yahei Mono, Microsoft YaHei'
        self.font_p_size = 15
        self.font = self.new_font()
        self.fm = QFontMetrics(self.font)
        self._char_height = self.fm.height()
        self._char_width = self.fm.width("W")
        self._columns, self._rows = self._pixel2pos(self.width(), self.height())
        self.cursor_x = 0
        self.cursor_y = 0
        self._selection = None
        # cache
        self.pens = {}
        self.brushes = {}
        self.default_brush = QBrush(self.theme["bg"])
        self.default_pen = QPen(self.theme["color"])
        # pixmap, all operation on this pixmap in receive thread,
        #         then invoke UI thread to paint this pixmap on main widget
        self.pixmap = QPixmap(self.width(), self.height())
        self.pixmap.fill(self.theme["bg"])
        # scroll
        self.scrollBar = None
        self.scrollEventSet = False
        self.scrollStatus = (0, 0)
        # timer for check change
        self.startTimer(100)
        self.updatePixmapLock = threading.Lock()
        # selection
        self.selectionStart = None
        self.lastSelection = None
        self.selectScroll = {     # for speed up scroll when select
            "speed": 0.1,
            "start_speed": 0.1,     # 0.5s
            "max_speed": 0.01,
            "ts": 0
        }

    def new_font(self):
        font = QFont()
        font.setFamily(self.font_name)
        font.setPixelSize(self.font_p_size)
        return font

    def get_pen(self, color_name="white", keyword = None):
        pen = self.pens.get(color_name)
        if not pen:
            if keyword:
                color = self.theme.get(keyword)
            else:
                color = self.theme["colors"].get(color_name)
            if not color:
                pen = self.default_pen
            else:
                pen = QPen(color)
            self.pens[color_name] = pen
        return pen

    def get_brush(self, color_name):
        brush = self.brushes.get(color_name)
        if not brush:
            color = self.theme["colors"].get(color_name)
            if not color:
                brush = self.default_brush
            else:
                brush = QBrush(color)
            self.brushes[color_name] = brush
        return brush

    def _pixel2pos(self, x, y):
        col = int(x / self._char_width)
        row = int(y / self._char_height)
        return col, row

    def _pos2pixel(self, col, row):
        x = col * self._char_width
        y = row * self._char_height
        return x, y

    def paint_full_pixmap(self):
        painter = QPainter(self.pixmap)
        screen = self.backend.screen
        self.paint_full_text(painter, screen)
        self.paint_cursor(painter, screen)

    def paint_part_pixmap(self):
        painter = QPainter(self.pixmap)
        screen = self.backend.screen
        self.paint_dirty_text(painter, screen)
        self.paint_cursor(painter, screen)

    def paint_full_text(self, painter, screen):
        painter.setFont(self.font)
        for line_num in range(self._rows):
            self.paint_line_text(painter, screen, line_num, clear=True)

    def paint_dirty_text(self, painter, screen):
        painter.setFont(self.font)
        # 重绘旧光标所在行
        screen.dirty.add(self.cursor_y)
        for line_num in screen.dirty:
            self.paint_line_text(painter, screen, line_num, clear=True)
        # clear paint dirty info mannually
        screen.dirty.clear()

    def paint_line_text(self, painter, screen, line_num, clear=False):
        start_x = 0
        start_y = line_num * self._char_height
        if clear:
            clear_rect = QRect(start_x, start_y, self.width(), self._char_height)
            painter.fillRect(clear_rect, self.default_brush)

        line = screen.buffer[line_num]

        same_text = ""
        text_width = 0
        pre_char = None
        align = Qt.AlignTop | Qt.AlignLeft

        for col in range(screen.columns):
            char = line[col]
            if pre_char and char.fg == pre_char.fg and char.bg == pre_char.bg:
                same_text += char.data
                continue
            else:
                if same_text:
                    text_width = self.fm.width(same_text)
                    self.draw_text(same_text, start_x, start_y, text_width, pre_char.fg, pre_char.bg, painter, align)

                pre_char = char
                same_text = char.data
                start_x = start_x + text_width

        if same_text:
            text_width = self.fm.width(same_text)
            self.draw_text(same_text, start_x, start_y, text_width, pre_char.fg, pre_char.bg, painter, align)

    def draw_text(self, text, start_x, start_y, text_width, fg, bg, painter, align):
        rect = QRect(start_x, start_y, text_width, self._char_height)

        if bg and bg != 'default':
            painter.fillRect(rect, self.get_brush(bg))

        painter.setPen(self.get_pen(fg))
        painter.drawText(rect, align, text)

    def paint_cursor(self, painter, screen):
        align = Qt.AlignTop | Qt.AlignLeft
        cursor = screen.cursor
        self.cursor_x = cursor.x
        self.cursor_y = cursor.y
        char = ""
        if self.cursor_y >= len(screen.display):
            cursor_x_pixel_width = 0
        else:
            line = screen.display[self.cursor_y]
            # to get actual text before cursor.x
            x = self.cursor_x
            while 1:
                w = wcswidth(line[:x])
                if w > self.cursor_x:
                    x -= 1
                elif w < self.cursor_x:
                    x += 1
                else:
                    break
            text = line[:x]
            if x < len(line):
                char = line[x]
            cursor_x_pixel_width = self.fm.width(text)
        bcol = self.theme["cursor"]
        brush = QBrush(bcol)

        rect = QRect(cursor_x_pixel_width, self.cursor_y * self._char_height,
                    self._char_width, self._char_height)
        painter.setPen(Qt.NoPen)
        painter.setBrush(brush)
        painter.drawRect(rect)
        if char:
            painter.setPen(self.get_pen(keyword="bg"))
            painter.drawText(rect, align, char)

    def wheelEvent(self, event: QWheelEvent):
        try:
            y = event.angleDelta().y()
            if y > 0:
                self.backend.screen.prev_page()
            else:
                self.backend.screen.next_page()
            self.update()
        except:
            import traceback
            traceback.print_exc()

    def onScrolled(self, curr):
        self.updatePixmapLock.acquire()
        top = len(self.backend.screen.history.top)
        bottom = len(self.backend.screen.history.bottom)
        maxlen = self.backend.screen.history.size
        if top + bottom + self._rows > maxlen:
            newBottom = maxlen - self._rows - curr
        else:
            newBottom = top + bottom - curr
        lastBottom = -1
        while 1:
            if newBottom > bottom:
                self.backend.screen.prev_page()
            elif newBottom < bottom:
                self.backend.screen.next_page()
            else:
                break
            # move up 1 line(ratio must set to move on line every prev_page)
            bottom = len(self.backend.screen.history.bottom)
            if bottom == lastBottom:
                break
            lastBottom = bottom
        self.updatePixmapLock.release()


    def paintEvent(self, event):
        '''
            paint pixmap to on widget
        '''
        painter = QPainter(self)
        painter.drawPixmap(0, 0, self.pixmap)
        # set scrollbar
        top = len(self.backend.screen.history.top)
        bottom = len(self.backend.screen.history.bottom)
        maxlen = self.backend.screen.history.size
        if top + bottom + self._rows > maxlen:
            barCount = maxlen - self._rows
            curr = maxlen - bottom - self._rows
        else:
            barCount = top + bottom
            curr = top
        # prevent scrollbar event
        if self.scrollStatus != (curr, barCount):
            if self.scrollEventSet:
                self.scrollBar.valueChanged.disconnect()
            self.scrollBar.setMaximum(barCount)
            self.scrollBar.setMinimum(0)
            self.scrollBar.setSliderPosition(curr)
            self.scrollBar.valueChanged.connect(self.onScrolled)
            self.scrollEventSet = True
            self.scrollStatus = (curr, barCount)

    def timerEvent(self, event):
        '''
            scheduled check change
        '''
        self.updatePixmapLock.acquire()
        self.updatePixmap()
        self.updatePixmapLock.release()

    def onReceived(self, data:bytes):
        '''
            in receive thread, not UI thread
            lock to make sure receive thread not change screen when check changes here
        '''
        self.updatePixmapLock.acquire()
        self.backend.feed(data)
        self.updatePixmap()
        self.updatePixmapLock.release()

    def updatePixmap(self):
        cursor = self.backend.screen.cursor
        # display changed or cursor changed
        if self.backend.screen.dirty or (self.cursor_x != cursor.x or self.cursor_y != cursor.y):
            self.paint_part_pixmap()
            # trigger paintEvent
            self.update()

    def event(self, e):
        if type(e) == QKeyEvent and e.key() == Qt.Key_Tab:
            if e.type() == QKeyEvent.KeyPress:
                self.keyPressEvent(e)
            return True
        return super().event(e)

    def inputMethodEvent(self, e):
        text = e.commitString()
        if text:
            self.send(text.encode("utf-8"))

    def keyPressEvent(self, event):
        text = str(event.text())
        key = event.key()
        modifiers = event.modifiers()
        ctrl = (modifiers == Qt.ControlModifier | Qt.ShiftModifier) or (modifiers == Qt.ControlModifier)
        if ctrl and key == Qt.Key_V:
            content = pyperclip.paste()
            self.send(content.encode("utf-8"))
        elif text and not key in self.keymap:
            self.send(text.encode("utf-8"))
        else:
            s = self.keymap.get(key)
            if s:
                self.send(s)
        event.accept()

    def resizeEvent(self, event):
        try:
            self.updatePixmapLock.acquire()
            self._columns, self._rows = self._pixel2pos(self.width(), self.height())
            self.backend.resize(self._columns, self._rows)
            self.resizeConn(self._columns, self._rows)
            self.pixmap = QPixmap(self.width(), self.height())
            self.pixmap.fill(self.theme["bg"])
            self.paint_full_pixmap()
            self.updatePixmapLock.release()
        except:
            import traceback
            traceback.print_exc()

    def setScrollBar(self, scrollBar):
        self.scrollBar = scrollBar
        self.scrollBar.setMinimum(0)
        height = len(self.backend.screen.history.top) + len(self.backend.screen.history.bottom) - self._rows
        self.scrollBar.setMaximum(height if height > 0 else 0)
        self.scrollBar.setMinimum(0)
        if not self.scrollEventSet:
            self.scrollBar.valueChanged.connect(self.onScrolled)
            self.scrollEventSet = True

    def _pos_pix2colrow(self, x, y):
        '''
            @return x, y, scroll(-1: up, 0:not scroll, 1: down)
        '''
        scroll = 0
        row = y // self._char_height
        if row >= self._rows:
            row = self._rows - 1
            scroll = 1
        if row < 0:
            row = 0
            col = 0
            scroll = -1
        else:
            col = round(x / self._char_width)
            if col > self._columns:
                col = self._columns
            elif col < 0:
                col = 0
        return col, row, scroll

    def _pos_to_absolute(self, col, row):
        row = len(self.backend.screen.history.top) + row
        return col, row

    def _copy2clipboard(self, start, end):
        content = self._get_selected_data(start, end)
        pyperclip.copy(content)

    def _get_selected_data(self, start, end):
        pixLen = 0
        content = ""
        if start[3] == end[3]: # same row
            line = self._get_line(start[3])
            for i in line:
                pixLen += self.fm.width(line[i].data)
                if pixLen > start[2] * self._char_width:
                    content += line[i].data
                if pixLen >= end[0]:
                    break
        else:
            # select last line to end
            line = self._get_line(start[3])
            draw = False
            for i in line:
                if not draw:
                    pixLen += self.fm.width(line[i].data)
                    if pixLen > start[2] * self._char_width:
                        draw = True
                    else:
                        continue
                if draw:
                    content += line[i].data
            content += "\n"
            # select middle lines(if have)
            for r in range(start[3] + 1, end[3]):
                line = self._get_line(r)
                for i in line:
                    content += line[i].data
                content += "\n"
            # select from the start of current line to the pos of pointer
            line = self._get_line(end[3])
            pixLen = 0
            for i in line:
                pixLen += self.fm.width(line[i].data)
                if pixLen <= end[2] * self._char_width:
                    content += line[i].data
                if pixLen >= end[0]:
                    break
        return content

    def mousePressEvent(self, e):
        self.mouseButton = e.button()
        x = e.x(); y = e.y()
        col, row, scroll = self._pos_pix2colrow(x, y)
        col, row = self._pos_to_absolute(col, row)
        self.selectionStart = (x, y, col, row)
        if self.lastSelection:
            if self.lastSelection[0] == self.lastSelection[1] and e.button() == Qt.RightButton: # no selection
                content = pyperclip.paste()
                self.send(content.encode("utf-8"))
            else:
                # mouse right button copy content to clipboard
                if e.button() == Qt.RightButton:
                    self._copy2clipboard(self.lastSelection[0], self.lastSelection[1])
                # unselect
                self.select(self.lastSelection[0], self.lastSelection[1], unselect=True)
        self.lastSelection = [self.selectionStart, self.selectionStart]
        self.selectScroll["speed"] =self.selectScroll["start_speed"]

    def toggleSelect(self, item):
        if item.bg == "selected":
            item = item._replace(bg="default")
        else:
            item = item._replace(bg="selected")
        return item

    def mouseMoveEvent(self, e):
        if self.mouseButton != Qt.LeftButton: # e.button() not work here, so use variable
            return
        self.updatePixmapLock.acquire()
        x = e.x(); y = e.y()
        col, row, scroll = self._pos_pix2colrow(x, y)
        col, row = self._pos_to_absolute(col, row)
        cursorPos = (x, y, col, row)
        # scroll
        if scroll != 0 and time.time() - self.selectScroll["ts"] > self.selectScroll["speed"]:
            if scroll < 0:
                self.backend.screen.prev_page()
            else:
                self.backend.screen.next_page()
            self.selectScroll["ts"] = time.time()
            if self.selectScroll["speed"] > self.selectScroll["max_speed"]:
                self.selectScroll["speed"] -= 0.005

        # get selection
        if col + row * self._columns > self.selectionStart[2] + self.selectionStart[3] * self._columns:
            front2rear = True
        else:
            front2rear = False
        if front2rear:
            newSelection = [self.selectionStart, cursorPos]
        else:
            newSelection = [cursorPos, self.selectionStart]
        # selection not change
        if newSelection[0][2:] == self.lastSelection[0][2:] and newSelection[1][2:] == self.lastSelection[1][2:]:
            self.updatePixmapLock.release()
            if scroll != 0:
                self.update()
            return
        # find need select range
        # find need erase election range
        needSlect = []; needErase = []
        newIdx = newSelection[0][2] + newSelection[0][3] * self._columns
        oldIdx = self.lastSelection[0][2] + self.lastSelection[0][3] * self._columns
        if newIdx < oldIdx:
            needSlect.append((newSelection[0], self.lastSelection[0]))
        elif newIdx > oldIdx:
            needErase.append((self.lastSelection[0], newSelection[0]))
        newIdx = newSelection[1][2] + newSelection[1][3] * self._columns
        oldIdx = self.lastSelection[1][2] + self.lastSelection[1][3] * self._columns
        if newIdx < oldIdx:
            needErase.append((newSelection[1], self.lastSelection[1]))
        elif newIdx > oldIdx:
            needSlect.append((self.lastSelection[1], newSelection[1]))
        # select
        for start, end in needSlect:
            self.select(start, end)
        for start, end in needErase:
            self.select(start, end, unselect = True)
        self.updatePixmapLock.release()
        # unselect
        # update last selection
        self.lastSelection = newSelection
        self.update()

    def mouseReleaseEvent(self, e):
        self.mouseButton = None

    def _get_line(self, row):
        if row < len(self.backend.screen.history.top):
            line = self.backend.screen.history.top[row]
        elif row < len(self.backend.screen.history.top) + self._rows:
            line = self.backend.screen.buffer[row - len(self.backend.screen.history.top)]
            self.backend.screen.dirty.add(row - len(self.backend.screen.history.top))
        else:
            line = self.backend.screen.history.bottom[row - len(self.backend.screen.history.top) - self._rows]
        return line

    def select(self, start, end, unselect = False):
        bg = "default" if unselect else "selected"
        pixLen = 0
        if start[3] == end[3]: # same row
            line = self._get_line(start[3])
            for i in line:
                pixLen += self.fm.width(line[i].data)
                if pixLen > start[2] * self._char_width:
                    line[i] = line[i]._replace(bg= bg)
                if pixLen >= end[0]:
                    break
        else:
            # select last line to end
            line = self._get_line(start[3])
            draw = False
            for i in line:
                if not draw:
                    pixLen += self.fm.width(line[i].data)
                    if pixLen > start[2] * self._char_width:
                        draw = True
                    else:
                        continue
                if draw:
                    line[i] = line[i]._replace(bg=bg)
            # select middle lines(if have)
            for r in range(start[3] + 1, end[3]):
                line = self._get_line(r)
                for i in line:
                    line[i] = line[i]._replace(bg=bg)
            # select from the start of current line to the pos of pointer
            line = self._get_line(end[3])
            pixLen = 0
            for i in line:
                pixLen += self.fm.width(line[i].data)
                if pixLen <= end[2] * self._char_width:
                    line[i] = line[i]._replace(bg=bg)
                if pixLen >= end[0]:
                    break


class Scroll_Terminal(QWidget):
    def __init__(self, sendFunc, resizeConnFunc, isConnectFunc) -> None:
        super().__init__()
        self.terminal = Terminal_Frontend(sendFunc, resizeConnFunc, isConnectFunc)
        self.scrollBar = QScrollBar(Qt.Vertical, self.terminal)
        self.terminal.setScrollBar(self.scrollBar)
        self.scrollBar.setProperty("class", "scrollbar2")
        layout = QHBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.terminal)
        layout.addWidget(self.scrollBar)
        self.resize(self.terminal.size())

    def onReceived(self, data):
        self.terminal.onReceived(data)

class Plugin(Plugin_Base):
    '''
        call sequence:
            set vars like hintSignal, hintSignal
            onInit
            onWidget
            onUiInitDone
                send
                onReceived
            onDel
    '''
    # vars set by caller
    isConnected = lambda : False
    send = lambda x,y:None          # send(data_bytes=None, file_path=None, callback=lambda ok,msg:None)
    hintSignal = None               # hintSignal.emit(type(error, warning, info), title, msg)
    reloadWindowSignal = None       # reloadWindowSignal.emit(title, msg, callback(close or not)), reload window to load new configs
    configGlobal = {}
    # other vars
    connParent = "main"      # parent id
    connChilds = []          # children ids
    id = "terminal"
    name = _("terminal")

    enabled = False          # user enabled this plugin
    active  = False          # using this plugin

    def __init__(self):
        super().__init__()
        if not self.id:
            raise ValueError(f"var id of Plugin {self} should be set")

    def onInit(self, config):
        '''
            init params, DO NOT take too long time in this func
            @config dict type, just change this var's content,
                               when program exit, this config will be auto save to config file
        '''
        self.config = config
        default = {
            "version": 1,
        }
        for k in default:
            if not k in self.config:
                self.config[k] = default[k]

    def onActive(self):
        pass

    def onDel(self):
        pass

    def onWidgetMain(self, parent):
        self.widget = Scroll_Terminal(self.send, self.resizeConnOutput, self.isConnected)
        return self.widget

    def resizeConnOutput(self, w, h):
        if self.isConnected:
            self.ctrlConn("resize", (w, h))

    def onReceived(self, data : bytes):
        self.widget.onReceived(data)
        for plugin in self.connChilds:
            plugin.onReceived(data)

    def onKeyPressEvent(self, event):
        pass

    def onKeyReleaseEvent(self, event):
        pass

    def onUiInitDone(self):
        '''
            UI init done, you can update your widget here
            this method runs in UI thread, do not block too long
        '''
        pass

    def onConnChanged(self, status:ConnectionStatus, msg:str):
        '''
            call in UI thread, be carefully!!
        '''
        super().onConnChanged(status, msg)
        if status == ConnectionStatus.CONNECTED:
            # set pty size when first connect
            self.widget.terminal.resizeEvent("")
            # auto focus terminal
            self.widget.terminal.setFocus()




import os
import gettext
import babel
from collections import OrderedDict

locales=["en", "zh_CN", "zh_TW", "ja"]


root_dir = os.path.abspath(os.path.dirname(__file__))
locale = "en"

tr = lambda x:x

def _(text):
    return tr(text)

def set_locale(locale_in):
    global locale, tr, root_dir
    print("-- set locale to", locale_in)
    locale = locale_in
    locales_path = os.path.join(root_dir, 'locales')
    if not os.path.exists(locales_path): # for pyinstaller pack
        locales_path = os.path.join(os.path.dirname(root_dir), 'locales')
    # check translate binary file
    mo_path = os.path.join(locales_path, "en", "LC_MESSAGES", "messages.mo")
    if not os.path.exists(mo_path):
        main("finish")
    lang = gettext.translation('messages', localedir=locales_path, languages=[locale])
    tr = lang.gettext

def get_languages():
    languages = OrderedDict()
    for locale in locales:
        obj = babel.Locale.parse(locale)
        languages[locale] = obj.language_name + (" " + obj.script_name if obj.script_name else "")
    return languages

def extract(src_path, config_file_path, out_path):
    from distutils.errors import DistutilsOptionError
    from babel.messages.frontend import extract_messages
    cmdinst = extract_messages()
    cmdinst.initialize_options()
    cmdinst.mapping_file = config_file_path
    cmdinst.output_file = out_path
    cmdinst.input_paths = src_path
    try:
        cmdinst.ensure_finalized()
        cmdinst.run()
    except DistutilsOptionError as err:
        raise err

def init(template_path, out_dir, locale, domain="messages"):
    from distutils.errors import DistutilsOptionError
    from babel.messages.frontend import init_catalog
    cmdinst = init_catalog()
    cmdinst.initialize_options()
    cmdinst.input_file = template_path
    cmdinst.output_dir = out_dir
    cmdinst.locale = locale
    cmdinst.domain = domain
    try:
        cmdinst.ensure_finalized()
        cmdinst.run()
    except DistutilsOptionError as err:
        raise err

def update(template_path, out_dir, locale, domain="messages"):
    from distutils.errors import DistutilsOptionError
    from babel.messages.frontend import update_catalog
    cmdinst = update_catalog()
    cmdinst.initialize_options()
    cmdinst.input_file = template_path
    cmdinst.output_dir = out_dir
    cmdinst.locale = locale
    cmdinst.domain = domain
    try:
        cmdinst.ensure_finalized()
        cmdinst.run()
    except DistutilsOptionError as err:
        raise err

def compile(translate_dir, locale, domain="messages"):
    from distutils.errors import DistutilsOptionError
    from babel.messages.frontend import compile_catalog
    cmdinst = compile_catalog()
    cmdinst.initialize_options()
    cmdinst.directory = translate_dir
    cmdinst.locale = locale
    cmdinst.domain = domain
    try:
        cmdinst.ensure_finalized()
        cmdinst.run()
    except DistutilsOptionError as err:
        raise err



def main(cmd, path=None):
    global root_dir
    babel_cfg_path = os.path.join(root_dir, "babel.cfg")
    if path:
        if os.path.exists(path):
            root_dir = os.path.abspath(path)
            if os.path.exists(os.path.join(root_dir, "babel.cfg")):
                babel_cfg_path = os.path.join(root_dir, "babel.cfg")
        else:
            print("path {} not exists".format(path))
            return

    cwd = os.getcwd()
    os.chdir(root_dir)
    if cmd == "prepare":
        print("== translate locales: {} ==".format(locales))
        print("-- extract keys from files")
        if not os.path.exists("locales"):
            os.makedirs("locales")
        # os.system("pybabel extract -F babel.cfg -o locales/messages.pot ./")
        extract("./", babel_cfg_path, "locales/messages.pot")
        for locale in locales:
            print("-- generate {} po files from pot files".format(locale))
            if os.path.exists('locales/{}/LC_MESSAGES/messages.po'.format(locale)):
                print("-- file already exits, only update")
                # "pybabel update -i locales/messages.pot -d locales -l {}".format(locale)
                update("locales/messages.pot", "locales", locale)
            else:
                print("-- file not exits, now create")
                # "pybabel init -i locales/messages.pot -d locales -l {}".format(locale)
                init("locales/messages.pot", "locales", locale)
    elif cmd == "finish":
        print("== translate locales: {} ==".format(locales))
        for locale in locales:
            print("-- generate {} mo file from po files".format(locale))
            # "pybabel compile -d locales -l {}".format(locale)
            compile("locales", locale)
    os.chdir(cwd)


def cli_main():
    import argparse
    parser = argparse.ArgumentParser("tranlate tool")
    parser.add_argument("-p", "--path", default="", help="path to the root of plugin")
    parser.add_argument("cmd", type=str, choices=["prepare", "finish"])
    args = parser.parse_args()
    main(args.cmd, args.path)

if __name__ == "__main__":
    cli_main()



import shutil
import sys, os
from datetime import datetime
import json

try:
    from i18n import _, set_locale
    from logger import Logger
except ImportError:
    from COMTool.i18n import _, set_locale
    from COMTool.logger import Logger

appName = "COMTool"
appIcon = "assets/logo.png"
appLogo = "assets/logo.png"
appLogo2 = "assets/logo2.png"
dataPath = os.path.abspath(os.path.dirname(__file__)).replace("\\", "/") # replace \ to / for qss usage, qss only support /
assetsDir = os.path.join(dataPath, "assets").replace("\\", "/")
if not os.path.exists(assetsDir): # for pyinstaller pack
    dataPath = os.path.dirname(dataPath)
    assetsDir = os.path.join(dataPath, "assets").replace("\\", "/")

defaultBaudrates = [9600, 19200, 38400, 57600, 74880, 115200, 921600, 1000000, 1500000, 2000000, 4500000]
encodings = ["ASCII", "UTF-8", "UTF-16", "GBK", "GB2312", "GB18030"]
customSendItemHeight = 40

author = "Neucrack"

def get_config_path(configFileName):
    configFilePath = configFileName
    try:
        configFilePath = os.path.join(configFileDir, configFileName)
        if not os.path.exists(configFileDir):
            os.makedirs(configFileDir)
    except:
        pass
    return configFilePath

configFileName="config.json"
configFilePath=configFileName

if sys.platform.startswith('linux') or sys.platform.startswith('darwin') or sys.platform.startswith('freebsd'):
    configFileDir = os.path.join(os.getenv("HOME"), ".config/comtool")
    configFilePath = get_config_path(configFileName)
else:
    configFileDir = os.path.abspath(os.getcwd())
    configFilePath  = os.path.join(configFileDir, configFileName)

logPath = os.path.join(configFileDir, "run.log")
log = Logger(file_path=logPath)
log.i("Config path:", configFilePath)
log.i("Log path:", logPath)


class Parameters:
    config = {
        "version": 3,
        "skin": "light",
        "locale": "en",
        "encoding": "UTF-8",
        "skipVersion": None,
        "connId": "serial",
        "pluginsInfo": {          # enabled plugins ID
            "external": {
                # "myplugin2": {
                #     # "package": "myplugin",  # package installed as a python package
                #     "path": "E:\main\projects\COMTool\COMTool\plugins\myplugin2\myplugin2.py"
                # }
            }
        },
        "activeItem": "dbg-1",
        "currItem": None,
        "items": [
            # {
            #     "name": "dbg-1",
            #     "pluginId": "dbg",
            #     "config": {
            #         "conns": {
            #             "currConn": "serial",
            #             "serial": {
            #              }
            #         },
            #         "plugin": {

            #         }
            #     }
            # }
        ]
    }

    def save(self, path):
        path = os.path.abspath(path)
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        with open(path, "w", encoding="utf-8") as f:
            json.dump(self.config, f, indent=4, ensure_ascii=False)

    def load(self, path):
        if not os.path.exists(path):
            return
        with open(path, encoding="utf-8") as f:
            config = json.load(f)
            if "version" in config and config["version"] == self.config["version"]:
                self.config = config
            else: # for old config, just backup
                t = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
                old_path = "{}.bak.{}.json".format(path, t)
                log.w("Old config file, backup to", old_path)
                shutil.copyfile(path, old_path)
        return

    def __getitem__(self, idx):
        return self.config[idx]

    def __setitem__(self, idx, v):
        self.config[idx] = v

    def __str__(self) -> str:
        return json.dumps(self.config)


strStyleShowHideButtonLeft = '''
QPushButton {
    border-image: url("$DataPath/assets/arrow-left.png")
}
QPushButton:hover {
    border-image: url("$DataPath/assets/arrow-left-white.png")
}'''

strStyleShowHideButtonRight = '''
QPushButton {
    border-image: url("$DataPath/assets/arrow-right.png")
}
QPushButton:hover {
    border-image: url("$DataPath/assets/arrow-right-white.png")
}'''

styleForCode = {
    "light":{
        "iconColor": "white",
        "iconSelectorColor": "#929599"
    },
    "dark":{
        "iconColor": "#bcbcbd",
        "iconSelectorColor": "#bcbcbd"
    }
}





from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import (QApplication, QWidget,QToolTip,QPushButton,QMessageBox,QDesktopWidget,QMainWindow,
                             QVBoxLayout,QHBoxLayout,QGridLayout,QTextEdit,QLabel,QRadioButton,QCheckBox,
                             QLineEdit,QGroupBox,QSplitter)


class Settings(QWidget):

    closed = pyqtSignal()
    updatedisTextRawSignal = pyqtSignal(str)
    buffer = ""

    def __init__(self,parent = None):
        super().__init__(parent)
        self.init()
        self.initEvent()
        self.show()

    def __del__(self):
        pass

    def init(self):
        self.resize(500,400)
        self.mainLayout = QVBoxLayout()
        self.setLayout(self.mainLayout)
        self.disTextRaw = QLabel("0000")
        self.mainLayout.addWidget(self.disTextRaw)

    def initEvent(self):
        pass

    def closeEvent(self, event):
        self.closed.emit()
        event.accept()







major = 3
minor = 3
dev   = 1

__version__ = "{}.{}.{}".format(major, minor, dev)

class Version:
    def __init__(self, major=major, minor=minor, dev=dev, name="", desc=""):
        self.major = major
        self.minor = minor
        self.dev   = dev
        self.name = name
        self.desc = desc

    def dump_dict(self):
        ret = {
            "major": self.major,
            "minor": self.minor,
            "dev": self.dev,
            "name": self.name,
            "desc": self.desc
        }
        return ret

    def load_dict(self, obj):
        self.major = obj['major']
        self.minor = obj['minor']
        self.dev= obj['dev']
        self.name = obj['name']
        self.desc = obj['desc']

    def int(self):
        return self.major * 100 + self.minor * 10 + self.dev

    def __str__(self):
        return 'v{}.{}.{}, {}: {}'.format(self.major, self.minor, self.dev, self.name, self.desc)



import imp
from PyQt5.QtCore import pyqtSignal, Qt, QRect, QMargins
from PyQt5.QtWidgets import (QApplication, QWidget,QPushButton,QMessageBox,QDesktopWidget,QMainWindow,
                             QVBoxLayout,QHBoxLayout,QGridLayout,QTextEdit,QLabel,QRadioButton,QCheckBox,
                             QLineEdit,QGroupBox,QSplitter,QFileDialog, QScrollArea, QTabWidget, QMenu, QSplashScreen)
from PyQt5.QtGui import QIcon,QFont,QTextCursor,QPixmap,QColor, QCloseEvent
import threading
import time
import os
import json

try:
    from i18n import _
    from Combobox import ComboBox
    from parameters import log, configFilePath
except ImportError:
    from COMTool.Combobox import ComboBox
    from COMTool.i18n import _
    from COMTool.parameters import log, configFilePath

class PluginItem:
    # display name
    name = ''
    widget = None
    def __init__(self, name, pluginClass,
                    connClasses, connsConfigs,
                    globalConfig, itemConfig,
                    hintSignal, reloadWindowSignal):
        '''
            item show name, e.g. dbg-1
        '''
        self.reloadWindowSignal = reloadWindowSignal
        self.hintSignal = hintSignal
        self.name = name
        self.connClasses = connClasses
        self.connsConfigs = connsConfigs
        self.currConnWidget = None
        self.currConnIdx = 0
        self.sendProcess = None
        self.dataToSend = []
        self.fileToSend = []
        # widgets
        self.settingWidget = None
        self.mainWidget = None
        self.functionalWidget = None
        # init plugin
        self.plugin = pluginClass()
        self.plugin.configGlobal = globalConfig
        self.plugin.send = self.sendData
        self.plugin.ctrlConn = self.ctrlConn
        self.plugin.hintSignal = self.hintSignal
        self.plugin.reloadWindowSignal = self.reloadWindowSignal
        self.plugin.onInit(config=itemConfig)
        if not "version" in itemConfig:
            raise Exception("{} {}".format(_("version not found in config of plugin:"), self.plugin.id))
        # conn
        self.conns, self.connWidgets = self.newConnWidgets()
        # frame
        self.widget = self.newFrame()
        self.uiLoadConfigs()
        self.initEvent()
    
    def newConnWidgets(self):
        connsWidgets = []
        conns = []
        for Conn in self.connClasses:
            # conn.onInit()
            conn = Conn()
            conns.append(conn)
            if conn.id in self.connsConfigs:
                connConfig = self.connsConfigs[conn.id]
            else:
                connConfig = {}
                self.connsConfigs[conn.id] = connConfig
            conn.onInit(connConfig)
            widget = conn.onWidget()
            conn.onUiInitDone()
            connsWidgets.append(widget)
        return conns, connsWidgets

    def newFrame(self):
        wrapper = QWidget()
        wrapperLayout = QVBoxLayout()
        wrapperLayout.setContentsMargins(0, 0, 0, 0)
        widget = QSplitter(Qt.Horizontal)
        widget.setProperty("class", "contentWrapper")
        statusBar = self.plugin.onWidgetStatusBar(wrapper)
        wrapper.setLayout(wrapperLayout)
        wrapperLayout.addWidget(widget)
        if not statusBar is None:
            wrapperLayout.addWidget(statusBar)
        # widgets settings
        self.settingWidget = QWidget()
        self.settingWidget.setProperty("class","settingWidget")
        settingLayout = QVBoxLayout()
        self.settingWidget.setLayout(settingLayout)
        #    get connection settings widgets
        connSettingsGroupBox = QGroupBox(_("Connection"))
        layout = QVBoxLayout()
        connSettingsGroupBox.setLayout(layout)
        self.connSelectCommbox = ComboBox()
        for conn in self.conns:
            self.connSelectCommbox.addItem(conn.name)
        layout.addWidget(self.connSelectCommbox)
        layout.setContentsMargins(1, 6, 0, 0)
        self.connsParent = QWidget()
        layout2 = QVBoxLayout()
        layout2.setContentsMargins(0, 0, 0, 0)
        self.connsParent.setLayout(layout2)
        layout.addWidget(self.connsParent)
        settingLayout.addWidget(connSettingsGroupBox)
        #    get settings widgets
        subSettingWidget = self.plugin.onWidgetSettings(widget)
        if not subSettingWidget is None:
            settingLayout.addWidget(subSettingWidget)
        settingLayout.addStretch()
        # widgets main
        self.mainWidget = self.plugin.onWidgetMain(widget)
        # widgets functional
        self.functionalWidget = QWidget()
        layout3 = QVBoxLayout()
        self.functionalWidget.setLayout(layout3)
        loadConfigBtn = QPushButton(_("Load config"))
        shareConfigBtn = QPushButton(_("Share config"))
        layout3.addWidget(loadConfigBtn)
        layout3.addWidget(shareConfigBtn)
        loadConfigBtn.clicked.connect(lambda : self.selectLoadfile())
        shareConfigBtn.clicked.connect(lambda : self.selectSharefile())
        pluginFuncWidget = self.plugin.onWidgetFunctional(widget)
        if not pluginFuncWidget is None:
            layout3.addWidget(pluginFuncWidget)
        layout3.addStretch()
        # add to frame
        widget.addWidget(self.settingWidget)
        widget.addWidget(self.mainWidget)
        widget.addWidget(self.functionalWidget)
        widget.setStretchFactor(0, 1)
        widget.setStretchFactor(1, 2)
        widget.setStretchFactor(2, 1)
        self.functionalWidget.hide()
        # UI init done
        self.plugin.onUiInitDone()
        return wrapper

    # event
    def selectSharefile(self):
        oldPath = os.getcwd()
        fileName_choose, filetype = QFileDialog.getSaveFileName(self.functionalWidget,
                            _("Select file"),
                            os.path.join(oldPath, f"comtool.{self.name}.json"),
                            _("json file (*.json);;config file (*.conf);;All Files (*)"))
        if fileName_choose != "":
            with open(fileName_choose, "w", encoding="utf-8") as f:
                for item in self.plugin.configGlobal["items"]:
                    if item["name"] == self.name:
                        json.dump(item, f, indent=4, ensure_ascii=False)
                        break

    def selectLoadfile(self):
        oldPath = os.getcwd()
        fileName_choose, filetype = QFileDialog.getOpenFileName(self.functionalWidget,
                                _("Select file"),
                                oldPath,
                                _("json file (*.json);;config file (*.conf);;All Files (*)"))
        if fileName_choose != "":
            with open(fileName_choose, "r", encoding="utf-8") as f:
                config = json.load( f)
                if "pluginsInfo" in config: # global config file
                    self.hintSignal.emit("error", _("Error"), _("Not support load global config file, you can copy config file mannually to " + configFilePath))
                    return
                if config["pluginId"] != self.plugin.id:
                    self.hintSignal.emit("error", _("Error"), _("Config is not for this plugin, config is for plugin:") + " " + config["pluginId"])
                    return
                if config["config"]["plugin"]["version"] != self.plugin.config["version"]:
                    self.hintSignal.emit("warning", _("Warning"), "{} {}, {}: {}, {}: {}".format(
                            _("Config version not same, plugin config version:"), config["config"]["plugin"]["version"],
                            _("now"), self.plugin.config["version"],
                            _("this maybe lead to some problem, if happened, please remove it manually from config file"),
                            configFilePath))
                    return
                self.oldConnConfigs = self.connsConfigs.copy()
                self.oldPluginConfigs = self.plugin.config.copy()
                self.connsConfigs.clear()
                self.plugin.config.clear()
                for k, v in config["config"]["conns"].items():
                    self.connsConfigs[k] = v
                for k, v in config["config"]["plugin"].items():
                    self.plugin.config[k] = v
                def onClose(ok):
                    if not ok:
                        self.connsConfigs.clear()
                        self.connsConfigs.update(self.oldConnConfigs)
                        self.plugin.config.clear()
                        self.plugin.config.update(self.oldPluginConfigs)
                self.reloadWindowSignal.emit("", _("Restart to load config?"), onClose)

    def _setConn(self, idx):
        if self.currConnWidget:
            self.currConnWidget.setParent(None)
            self.conns[self.currConnIdx].onReceived = lambda x:None
            self.conns[self.currConnIdx].onConnectionStatus.disconnect(self.onConnStatus)
        self.currConnWidget = self.connWidgets[idx]
        self.connsParent.layout().addWidget(self.currConnWidget)
        self.conns[idx].onReceived = self.onReceived
        self.plugin.isConnected = self.conns[idx].isConnected
        self.conns[idx].onConnectionStatus.connect(self.onConnStatus)
        self.connsConfigs["currConn"] = self.conns[idx].id
        self.currConnIdx = idx


    def uiLoadConfigs(self):
        loadedIdx = 0
        if "currConn" in self.connsConfigs:
            for idx, conn in enumerate(self.conns):
                if conn.id == self.connsConfigs["currConn"]:
                    loadedIdx = idx
        self.connSelectCommbox.setCurrentIndex(loadedIdx)
        self._setConn(loadedIdx)

    def initEvent(self):
        self.connSelectCommbox.currentIndexChanged.connect(self.onConnChanged)

    def onConnChanged(self, idx):
        self.conns[self.currConnIdx].disconnect()
        self._setConn(idx)

    def ctrlConn(self, k, v):
        self.conns[self.currConnIdx].ctrl(k, v)

    def onConnStatus(self, status, msg):
        self.plugin.onConnChanged(status, msg)
        if self.sendProcess is None:
            self.sendProcess = threading.Thread(target=self.sendDataProcess)
            self.sendProcess.setDaemon(True)
            self.sendProcess.start()

    def sendData(self, data_bytes=None, file_path=None, callback=lambda ok, msg, length, path:None):
        if data_bytes:
            self.dataToSend.insert(0, (data_bytes, callback))
        if file_path:
            self.fileToSend.insert(0, (file_path, callback))

    def onReceived(self, data):
        self.plugin.onReceived(data)

    def onKeyPressEvent(self, e):
        self.plugin.onKeyPressEvent(e)

    def onKeyReleaseEvent(self, e):
        self.plugin.onKeyReleaseEvent(e)

    def sendDataProcess(self):
        self.receiveProgressStop = False
        while not self.receiveProgressStop:
            try:
                if not self.conns[self.currConnIdx].isConnected():
                    time.sleep(0.001)
                    continue
                while len(self.dataToSend) > 0:
                    data, callback = self.dataToSend.pop()
                    self.conns[self.currConnIdx].send(data)
                    callback(True, "", len(data), "")
                while len(self.fileToSend) > 0:
                    file_path, callback = self.fileToSend.pop()
                    ok = False
                    length = 0
                    if file_path and os.path.exists(file_path):
                        data = None
                        try:
                            with open(file_path, "rb") as f:
                                data = f.read()
                        except Exception as e:
                            self.hintSignal.emit("error", _("Error"), _("Open file failed!") + "\n%s\n%s" %(file_path, str(e)))
                        if data:
                            self.conns[self.currConnIdx].send(data)
                            length = len(data)
                            ok = True
                    callback(ok, "", length, file_path)
                time.sleep(0.001)
            except Exception as e:
                import traceback
                exc = traceback.format_exc()
                log.e(exc)
                if 'multiple access' in str(e):
                    self.hintSignal.emit("error", _("Error"), "device disconnected or multiple access on port?")
                continue

    def onDel(self):
        for conn in self.conns:
            conn.onDel()
        self.plugin.onDel()




from PyQt5.QtWidgets import (QWidget, QVBoxLayout, QLabel, QSlider, QLineEdit,
                             QGridLayout, QPushButton, QCheckBox, QHBoxLayout, QInputDialog, QComboBox, QFileDialog)
from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QDoubleValidator
import pyqtgraph as pg
import os
import numpy as np
from multiprocessing import Queue
from PIL import Image
# big package, DO NOT USE IT
# import matplotlib.pyplot as plt

from struct import unpack, pack
from .graph_widgets_base import Graph_Widget_Base
try:
    import utils
    from i18n import _
except Exception:
    from COMTool.i18n import _
    from COMTool import utils


# class Graph_DragTof(Graph_Widget_Base):
#     updateSignal = pyqtSignal(dict)
#     id = "DragTof"

#     def __init__(self, parent=None, hintSignal=lambda type, title, msg: None, rmCallback=lambda widget: None, send=lambda x: None, config=None):
#         default = {
#             "xRange": 10,
#             "xRangeEnable": True,
#             "header": "\\xCC\\xA0"
#         }
#         super().__init__(parent, hintSignal=hintSignal, rmCallback=rmCallback,
#                          send=send, config=config, defaultConfig=default)
#         self.headerBytes = utils.str_to_bytes(
#             self.config["header"], escape=True, encoding="utf-8")
#         self.layout = QGridLayout()
#         self.setLayout(self.layout)
#         self.imv = pg.ImageView()
#         self.queue = Queue()
#         # self.plotWin = pg.GraphicsLayoutWidget()
#         # self.plotWin.setMinimumHeight(200)
#         # self.plotWin
#         pg.setConfigOptions(antialias=True)
#         rmBtn = QPushButton(_("Remove"))
#         rangeLabel = QLabel(_("Range:"))
#         rangeConf = QLineEdit(str(self.config["xRange"]))
#         rangeEnable = QCheckBox(_("Enable"))
#         rangeEnable.setChecked(self.config["xRangeEnable"])
#         headerLabel = QLabel(_("Header:"))
#         headerConf = QLineEdit(self.config["header"])
#         self.headerBtn = QPushButton(_("Set"))
#         hint = _("Protocol: 2Byte header + 2Byte length + 1Byte command + 1Byte output_mode + 1Byte Sensor Temp + 1Byte Driver Temp\n"
#                  " 4Bytes exposure time + 1Byte error code + 1Byte reserved1 + 1Byte res rows + 1Byte res cols\n"
#                  " 2Byte Frame ID + 1 Byte ISP version + 1 Byte reserved3\n"
#                  " frame data\n"
#                  " 1Byte checksum + 1Byte tail\n"
#                  " length count from Byte4 to the Byte before Checksum\n"
#                  "Protocol example code see help")
#         headerConf.setToolTip(hint)
#         headerLabel.setToolTip(hint)
#         self.headerBtn.setToolTip(hint)
#         validator = QDoubleValidator()
#         rangeConf.setValidator(validator)
#         self.layout.addWidget(self.imv, 0, 0, 1, 3)
#         # self.layout.addWidget(self.plotWin, 0, 0, 1, 3)
#         self.layout.addWidget(rmBtn, 1, 0, 1, 1)
#         self.layout.addWidget(rangeLabel, 2, 0, 1, 1)
#         self.layout.addWidget(rangeConf, 2, 1, 1, 1)
#         self.layout.addWidget(rangeEnable, 2, 2, 1, 1)
#         self.layout.addWidget(headerLabel, 3, 0, 1, 1)
#         self.layout.addWidget(headerConf, 3, 1, 1, 1)
#         self.layout.addWidget(self.headerBtn, 3, 2, 1, 1)
#         self.resize(600, 400)
#         # self.p = self.plotWin.addPlot(colspan=2)
#         # self.p.addLegend()
#         # self.p.setXRange(0, self.config["xRange"])
#         self.updateSignal.connect(self.update)
#         self.rawData = b''
#         self.data = {}
#         self.builtinColors = [
#             "#BD4B4B",
#             "#3BB273",
#             "#FFFFFA",
#             "#307473",
#             "#3C6997",
#             "#746D75",
#             "#228CDB",
#             "#824C71",
#             "#7768AE",
#             "#DC6BAD",
#             "#607d8b",
#             "#F18701",
#             "#912F40",
#             "#414288",
#             "#ED4D6E",
#             "#FFD29D",
#             "#B56576",
#             "#503B31",
#             "#93E1D8",
#             "#596157",
#         ]
#         self.notUsedColors = self.builtinColors.copy()
#         self.colors = {

#         }
#         rangeConf.textChanged.connect(self.setRange)
#         rangeEnable.clicked.connect(
#             lambda: self.setEnableRange(rangeEnable.isChecked()))
#         self.headerBtn.clicked.connect(
#             lambda: self.setHeader(headerConf.text()))
#         rmBtn.clicked.connect(self.remove)
#         headerConf.textChanged.connect(self.headerChanged)

#     def remove(self):
#         self.rmCallback(self)

#     def setRange(self, text):
#         if text:
#             self.config["xRange"] = float(text)

#     def setEnableRange(self, en):
#         self.config["xRangeEnable"] = en

#     def headerChanged(self, text):
#         if self.config["header"] != text:
#             self.headerBtn.setText(_("Set") + " *")
#         else:
#             self.headerBtn.setText(_("Set"))

#     def setHeader(self, text):
#         if text:
#             try:
#                 textBytes = utils.str_to_bytes(
#                     text, escape=True, encoding="utf-8")
#                 self.config["header"] = text
#                 self.headerBytes = textBytes
#                 self.headerBtn.setText(_("Set"))
#             except Exception:
#                 self.hintSignal.emit("error", _("Error"), _("Format error"))

#     def decodeData(self, data: bytes):
#         '''
#             @data bytes, protocol:
#                          |    header(2B)    | len(2B) | command(1B) | output_mode(1B)     | sensor temp(1B) | driver temp(1B) |
#                          |     0xCC  0xA0      |  >=18   |             | 0:Depth only, 1:+IR |    uint8        | uint8           |
#                          -------------------------------------------------------------------------------------------------------
#                          |    exposure time(4B)    | errcode(1B) | reserved1(1B) | res_rows(1B) | res_cols(1B) |
#                          |[23:20][19:12][11:8][7:0]|             |       0x0     |     uint8    |     uint8    |
#                          ---------------------------------------------------------------------------------------
#                          |    Frame ID(2B)    | ISP ver(1B) | reserved3(1B) | frame data((len-16)B) | checksum(1B) | tail(1B) |
#                          |     [11:0]         |     0x23    |       0x0     | xxxxxxxxxxxxxxxxxxxxx | sum of above |   0xDD   |

#             @return haveFrame, dict {
#                 "frameID": {
#                     "res": tunple(w, h)
#                     "frameData": []
#                 }
#             }
#         '''
#         # append data
#         self.rawData += data
#         # find header
#         header = self.headerBytes
#         idx = self.rawData.find(header)
#         if idx < 0:
#             return False, self.data
#         self.rawData = self.rawData[idx:]
#         # print(self.rawData)
#         # check data length 2Byte
#         dataLen = unpack("H", self.rawData[2: 4])[0]
#         # print(dataLen)
#         frameLen = len(header) + 2 + dataLen + 2
#         frameDataLen = dataLen - 16

#         if len(self.rawData) < frameLen:
#             return False, self.data
#         # get data
#         frame = self.rawData[:frameLen]
#         # print(frame)
#         self.rawData = self.rawData[frameLen:]

#         frameTail = frame[-1]
#         # print(frameTail)
#         _sum = frame[-2]
#         # print(_sum)
#         # check sum
#         if _sum != sum(frame[:frameLen - 2]) % 256:
#             return True, self.data

#         frameID = unpack("H", frame[16:18])[0]
#         # print(frameID)

#         resR = unpack("B", frame[14:15])[0]
#         resC = unpack("B", frame[15:16])[0]
#         res = (resR, resC)
#         # print(res)
#         frameData = [unpack("H", frame[20+i:22+i])[0]
#                      for i in range(0, frameDataLen, 2)]

#         # if not frameID in self.data:
#         #     self.data[frameID] = {
#         #         "res": res,
#         #         "frameData": frameData
#         #     }
#         # else:
#         #     self.data[frameID]["res"] = res
#         #     self.data[frameID]["frameData"] = frameData

#         queueData = {
#             'frameID': frameID,
#             "res": res,
#             "frameData": frameData
#         }
#         self.queue.put(queueData)

#         return True, self.data

#     def pickColor(self, name: str):
#         if name in self.colors:
#             return self.colors[name]
#         else:
#             if not self.notUsedColors:
#                 self.notUsedColors = self.builtinColors.copy()
#             color = self.notUsedColors.pop(0)
#             self.colors[name] = color
#             return color

#     def update(self, data: dict):
#         # for k, v in data.items():
#         if not self.queue.empty():
#             v = self.queue.get()
#             k = v['frameID']
#             print("draw frame:", k)
#             res = v['res']
#             frameData = v['frameData']
#             arr = np.array(frameData, np.uint16)
#             arr[np.where(arr > 3000)] = 0
#             arr = arr.reshape(res)
#             arr = np.expand_dims(arr, 2)
#             # print(v)
#             self.imv.setImage(arr)
#             # color = self.pickColor(k)
#             # self.colors[k] = color
#             # self.curves[k] = self.p.plot(pen=pg.mkPen(color=color, width=2),
#             #                             name=k,)
#             # self.curves[k].setData(x=v["x"], y=v["y"])

#     def onData(self, data: bytes):
#         while 1:
#             haveFrame, allData = self.decodeData(data)
#             if not haveFrame:
#                 break
#             data = b''
#         self.updateSignal.emit(allData)


class Graph_MetaSenseLite(Graph_Widget_Base):
    updateSignal = pyqtSignal(dict)
    id = "MetaSenseLite"

    def __init__(self, parent=None, hintSignal=lambda type, title, msg: None, rmCallback=lambda widget: None, send=lambda x: None, config=None):
        default = {
            "xRange": 10,
            "xRangeEnable": True,
            "header": "\\x00\\xFF"
        }
        super().__init__(parent, hintSignal=hintSignal, rmCallback=rmCallback,
                         send=send, config=config, defaultConfig=default)
        self.headerBytes = utils.str_to_bytes(
            self.config["header"], escape=True, encoding="utf-8")
        self.layout = QGridLayout()
        self.setLayout(self.layout)

        class ImageItemWithMouseHover(pg.ImageItem):
            def __init__(self, image=None, **kargs):
                super().__init__(image=image, kargs=kargs)

        pg.setConfigOptions(imageAxisOrder='row-major')
        self.imv = pg.ImageView(view=pg.PlotItem())
        self.imv.setMouseTracking(True)
        self.setMouseTracking(True)
        self.queue = Queue()
        self.onceShot = False
        self.frames = []
        self.distances = []
        jetcolors = [(128, 0, 0), (132, 0, 0), (136, 0, 0), (140, 0, 0), (144, 0, 0), (148, 0, 0), (152, 0, 0), (156, 0, 0), (160, 0, 0), (164, 0, 0), (168, 0, 0), (172, 0, 0), (176, 0, 0), (180, 0, 0), (184, 0, 0), (188, 0, 0), (192, 0, 0), (196, 0, 0), (200, 0, 0), (204, 0, 0), (208, 0, 0), (212, 0, 0), (216, 0, 0), (220, 0, 0), (224, 0, 0), (228, 0, 0), (232, 0, 0), (236, 0, 0), (240, 0, 0), (244, 0, 0), (248, 0, 0), (252, 0, 0), (255, 0, 0), (255, 4, 0), (255, 8, 0), (255, 12, 0), (255, 16, 0), (255, 20, 0), (255, 24, 0), (255, 28, 0), (255, 32, 0), (255, 36, 0), (255, 40, 0), (255, 44, 0), (255, 48, 0), (255, 52, 0), (255, 56, 0), (255, 60, 0), (255, 64, 0), (255, 68, 0), (255, 72, 0), (255, 76, 0), (255, 80, 0), (255, 84, 0), (255, 88, 0), (255, 92, 0), (255, 96, 0), (255, 100, 0), (255, 104, 0), (255, 108, 0), (255, 112, 0), (255, 116, 0), (255, 120, 0), (255, 124, 0), (255, 128, 0), (255, 132, 0), (255, 136, 0), (255, 140, 0), (255, 144, 0), (255, 148, 0), (255, 152, 0), (255, 156, 0), (255, 160, 0), (255, 164, 0), (255, 168, 0), (255, 172, 0), (255, 176, 0), (255, 180, 0), (255, 184, 0), (255, 188, 0), (255, 192, 0), (255, 196, 0), (255, 200, 0), (255, 204, 0), (255, 208, 0), (255, 212, 0), (255, 216, 0), (255, 220, 0), (255, 224, 0), (255, 228, 0), (255, 232, 0), (255, 236, 0), (255, 240, 0), (255, 244, 0), (255, 248, 0), (255, 252, 0), (254, 255, 1), (250, 255, 6), (246, 255, 10), (242, 255, 14), (238, 255, 18), (234, 255, 22), (230, 255, 26), (226, 255, 30), (222, 255, 34), (218, 255, 38), (214, 255, 42), (210, 255, 46), (206, 255, 50), (202, 255, 54), (198, 255, 58), (194, 255, 62), (190, 255, 66), (186, 255, 70), (182, 255, 74), (178, 255, 78), (174, 255, 82), (170, 255, 86), (166, 255, 90), (162, 255, 94), (158, 255, 98), (154, 255, 102), (150, 255, 106), (146, 255, 110), (142, 255, 114), (138, 255, 118), (134, 255, 122), (130, 255, 126),
                     (126, 255, 130), (122, 255, 134), (118, 255, 138), (114, 255, 142), (110, 255, 146), (106, 255, 150), (102, 255, 154), (98, 255, 158), (94, 255, 162), (90, 255, 166), (86, 255, 170), (82, 255, 174), (78, 255, 178), (74, 255, 182), (70, 255, 186), (66, 255, 190), (62, 255, 194), (58, 255, 198), (54, 255, 202), (50, 255, 206), (46, 255, 210), (42, 255, 214), (38, 255, 218), (34, 255, 222), (30, 255, 226), (26, 255, 230), (22, 255, 234), (18, 255, 238), (14, 255, 242), (10, 255, 246), (6, 255, 250), (2, 255, 254), (0, 252, 255), (0, 248, 255), (0, 244, 255), (0, 240, 255), (0, 236, 255), (0, 232, 255), (0, 228, 255), (0, 224, 255), (0, 220, 255), (0, 216, 255), (0, 212, 255), (0, 208, 255), (0, 204, 255), (0, 200, 255), (0, 196, 255), (0, 192, 255), (0, 188, 255), (0, 184, 255), (0, 180, 255), (0, 176, 255), (0, 172, 255), (0, 168, 255), (0, 164, 255), (0, 160, 255), (0, 156, 255), (0, 152, 255), (0, 148, 255), (0, 144, 255), (0, 140, 255), (0, 136, 255), (0, 132, 255), (0, 128, 255), (0, 124, 255), (0, 120, 255), (0, 116, 255), (0, 112, 255), (0, 108, 255), (0, 104, 255), (0, 100, 255), (0, 96, 255), (0, 92, 255), (0, 88, 255), (0, 84, 255), (0, 80, 255), (0, 76, 255), (0, 72, 255), (0, 68, 255), (0, 64, 255), (0, 60, 255), (0, 56, 255), (0, 52, 255), (0, 48, 255), (0, 44, 255), (0, 40, 255), (0, 36, 255), (0, 32, 255), (0, 28, 255), (0, 24, 255), (0, 20, 255), (0, 16, 255), (0, 12, 255), (0, 8, 255), (0, 4, 255), (0, 0, 255), (0, 0, 252), (0, 0, 248), (0, 0, 244), (0, 0, 240), (0, 0, 236), (0, 0, 232), (0, 0, 228), (0, 0, 224), (0, 0, 220), (0, 0, 216), (0, 0, 212), (0, 0, 208), (0, 0, 204), (0, 0, 200), (0, 0, 196), (0, 0, 192), (0, 0, 188), (0, 0, 184), (0, 0, 180), (0, 0, 176), (0, 0, 172), (0, 0, 168), (0, 0, 164), (0, 0, 160), (0, 0, 156), (0, 0, 152), (0, 0, 148), (0, 0, 144), (0, 0, 140), (0, 0, 136), (0, 0, 132), (0, 0, 128)]
        self.imv.setColorMap(pg.ColorMap(
            pos=np.linspace(0.0, 1.0, len(jetcolors)), color=jetcolors))
        # self.imv.setColorMap(pg.colormap.get('jet_r', source='matplotlib'))
        # self.imv.setColorMap(pg.colormap.get("CET-L17"))
        # self.plotWin = pg.GraphicsLayoutWidget()
        # self.plotWin.setMinimumHeight(200)
        # self.plotWin
        pg.setConfigOptions(antialias=True)
        rmBtn = QPushButton(_("Remove"))
        self.capTargetBtn = QPushButton(os.path.curdir)
        capBtn = QPushButton(_("Capture"))
        # rangeLabel = QLabel(_("Range:"))
        # rangeConf = QLineEdit(str(self.config["xRange"]))
        # rangeEnable = QCheckBox(_("Enable"))
        # rangeEnable.setChecked(self.config["xRangeEnable"])
        headerLabel = QLabel(_("Header:"))
        headerConf = QLineEdit(self.config["header"])
        self.headerBtn = QPushButton(_("Set"))

        rawCmdLabel = QLabel(_("RawCMD:"))
        rawCmdConf = QLineEdit(_("AT+ISP=0\\r"))
        rawCmdBtn = QPushButton(_("Send"))

        btnWin = QWidget()
        btnWin.layout = QVBoxLayout()
        btnWin.setLayout(btnWin.layout)
        ispRun = QCheckBox(_("ISP"))
        dispCbLcd = QCheckBox(_("LCD"))
        dispCbUsb = QCheckBox(_("USB"))
        dispCbUart = QCheckBox(_("UART"))
        antiMmiCb = QCheckBox(_("ANTIMMI"))

        vBoxWin = QWidget()
        vBoxWin.layout = QVBoxLayout()
        vBoxWin.setLayout(vBoxWin.layout)

        itemCb = QWidget()
        itemCb.layout = QHBoxLayout()
        itemCb.setLayout(itemCb.layout)

        binnItem = QComboBox()
        baudItem = QComboBox()

        focusCb = QCheckBox()
        focusX = QLineEdit("0")
        focusY = QLineEdit("0")
        focusD = QLineEdit("0")
        self.focus = (focusCb, (focusX, focusY), focusD)

        unitSliderBox = QWidget()
        unitSliderBox.layout = QHBoxLayout()
        unitSliderBox.setLayout(unitSliderBox.layout)
        unitLabel = QLabel(_("Unit:auto "))
        unitSlider = QSlider(Qt.Horizontal)
        self.unitSlider = unitSlider

        fpsSliderBox = QWidget()
        fpsSliderBox.layout = QHBoxLayout()
        fpsSliderBox.setLayout(fpsSliderBox.layout)
        fpsLabel = QLabel(_(" Fps:15   "))
        fpsSlider = QSlider(Qt.Horizontal)

        evSliderBox = QWidget()
        evSliderBox.layout = QHBoxLayout()
        evSliderBox.setLayout(evSliderBox.layout)
        evLabel = QLabel(_("  Ev:AE   "))
        evSlider = QSlider(Qt.Horizontal)

        hint = _("Protocol: 2Byte header + 2Byte length + 1Byte command + 1Byte output_mode + 1Byte Sensor Temp + 1Byte Driver Temp\n"
                 " 4Bytes exposure time + 1Byte error code + 1Byte reserved1 + 1Byte res rows + 1Byte res cols\n"
                 " 2Byte Frame ID + 1 Byte ISP version + 1 Byte reserved3\n"
                 " frame data\n"
                 " 1Byte checksum + 1Byte tail\n"
                 " length count from Byte4 to the Byte before Checksum\n"
                 "Protocol example code see help")
        headerConf.setToolTip(hint)
        headerLabel.setToolTip(hint)
        self.headerBtn.setToolTip(hint)
        # validator = QDoubleValidator()
        # rangeConf.setValidator(validator)
        self.layout.addWidget(self.imv, 0, 0, 3, 3)
        # self.layout.addWidget(self.plotWin, 0, 0, 1, 3)
        self.layout.addWidget(rmBtn, 3, 0, 1, 1)
        self.layout.addWidget(self.capTargetBtn, 3, 1, 1, 1)
        self.layout.addWidget(capBtn, 3, 2, 1, 1)
        # self.layout.addWidget(rangeLabel, 2, 0, 1, 1)
        # self.layout.addWidget(rangeConf, 2, 1, 1, 1)
        # self.layout.addWidget(rangeEnable, 2, 2, 1, 1)
        self.layout.addWidget(headerLabel, 4, 0, 1, 1)
        self.layout.addWidget(headerConf, 4, 1, 1, 1)
        self.layout.addWidget(self.headerBtn, 4, 2, 1, 1)

        self.layout.addWidget(rawCmdLabel, 5, 0, 1, 1)
        self.layout.addWidget(rawCmdConf, 5, 1, 1, 1)
        self.layout.addWidget(rawCmdBtn, 5, 2, 1, 1)

        self.layout.addWidget(btnWin, 6, 0, 2, 1)
        btnWin.layout.addWidget(ispRun)
        btnWin.layout.addWidget(dispCbLcd)
        btnWin.layout.addWidget(dispCbUsb)
        btnWin.layout.addWidget(dispCbUart)
        btnWin.layout.addWidget(antiMmiCb)

        self.layout.addWidget(vBoxWin, 6, 1, 2, 2)
        vBoxWin.layout.addWidget(itemCb)
        itemCb.layout.addWidget(QLabel(_("Binn:")))
        itemCb.layout.addWidget(binnItem)
        itemCb.layout.addWidget(QLabel(_("Baud:")))
        itemCb.layout.addWidget(baudItem)
        itemCb.layout.addWidget(focusCb)
        itemCb.layout.addWidget(QLabel(_("D:")))
        itemCb.layout.addWidget(focusD)
        itemCb.layout.addWidget(QLabel(_("X:")))
        itemCb.layout.addWidget(focusX)
        itemCb.layout.addWidget(QLabel(_("Y:")))
        itemCb.layout.addWidget(focusY)

        unitSliderBox.layout.addWidget(unitLabel)
        unitSliderBox.layout.addWidget(unitSlider)
        vBoxWin.layout.addWidget(unitSliderBox)

        fpsSliderBox.layout.addWidget(fpsLabel)
        fpsSliderBox.layout.addWidget(fpsSlider)
        vBoxWin.layout.addWidget(fpsSliderBox)

        evSliderBox.layout.addWidget(evLabel)
        evSliderBox.layout.addWidget(evSlider)
        vBoxWin.layout.addWidget(evSliderBox)

        self.resize(600, 400)
        # self.p = self.plotWin.addPlot(colspan=2)
        # self.p.addLegend()
        # self.p.setXRange(0, self.config["xRange"])
        self.updateSignal.connect(self.update)
        self.rawData = b''
        self.data = {}
        self.builtinColors = [
            "#BD4B4B",
            "#3BB273",
            "#FFFFFA",
            "#307473",
            "#3C6997",
            "#746D75",
            "#228CDB",
            "#824C71",
            "#7768AE",
            "#DC6BAD",
            "#607d8b",
            "#F18701",
            "#912F40",
            "#414288",
            "#ED4D6E",
            "#FFD29D",
            "#B56576",
            "#503B31",
            "#93E1D8",
            "#596157",
        ]
        self.notUsedColors = self.builtinColors.copy()
        self.colors = {

        }

        imgView = self.imv.getView()
        imgItem = self.imv.getImageItem()
        # p = imgItem.ItemCursorHasChanged.connect(lambda x: print(x))
        scene = imgItem.scene()

        # imgItem.mouseMoveEvent =

        # def ptint_wr(wr):
        #     if len(wr[0]) > 1:
        #         ichild = self.imv.children()[1].children()[0].children()[3]
        #         # print(ichild.pos())
        #         imgItm = wr[0][1]
        #         cursor_pos = imgItm.cursor().pos()

        #         ise = imgItm.scene()
        #         iv = ise.views()[0]
        #         print(iv.pos(), end=', ')
        #         abs_pos = iv.mapToGlobal(iv.pos())
        #         print(abs_pos)

        # self.proxy = pg.SignalProxy(scene.sigMouseHover, rateLimit=60,
        #                             slot=ptint_wr)

        # self.proxy1 = pg.SignalProxy(scene.sigMouseMoved , rateLimit=60,
        #                             slot=lambda x:print(x))

        # rangeConf.textChanged.connect(self.setRange)
        # rangeEnable.clicked.connect(lambda: self.setEnableRange(rangeEnable.isChecked()))
        self.headerBtn.clicked.connect(
            lambda: self.setHeader(headerConf.text()))
        headerConf.textChanged.connect(self.headerChanged)

        rmBtn.clicked.connect(self.remove)
        self.capTargetBtn.clicked.connect(lambda: self.capTargetBtn.setText(QFileDialog.getExistingDirectory(
            None, "Choose Directory To Save", os.path.expanduser("~")) or self.capTargetBtn.text()))
        capBtn.clicked.connect(self.capture)
        rawCmdBtn.clicked.connect(lambda: self.sendCmd(rawCmdConf.text()))

        ispRun.setChecked(True)
        ispRun.toggled.connect(
            lambda x: self.sendCmd("AT+ISP=%1d\r" % (x)))

        def dispChanged():
            self.sendCmd("AT+DISP=%1d\r" % (1 * dispCbLcd.isChecked() +
                         2 * dispCbUsb.isChecked() + 4 * dispCbUart.isChecked()))
        dispCbLcd.setChecked(True)
        dispCbLcd.toggled.connect(dispChanged)
        dispCbUsb.toggled.connect(dispChanged)
        dispCbUart.toggled.connect(dispChanged)

        antiMmiCb.setChecked(True)
        antiMmiCb.toggled.connect(
            lambda x: self.sendCmd("AT+ANTIMMI=%1d\r" % (x)))

        binnItem.addItems(["1x1", "2x2", "4x4"])
        binnItem.setCurrentIndex(0)
        binnItem.currentIndexChanged.connect(
            lambda i: self.sendCmd("AT+BINN=%1d\r" % (1 << i)))

        baudItem.addItems(["9600", " 57600", "115200", "230400",
                          "460800", "921600", "1000000", "2000000", "3000000"])
        baudItem.setCurrentIndex(2)
        baudItem.currentIndexChanged.connect(
            lambda i: self.sendCmd("AT+BAUD=%1d\r" % (i)))

        def focusCbChanged(sta):
            if sta:
                x, y = 0, 0
                try:
                    x = int(focusX.text())
                    y = int(focusY.text())
                    assert(x >= 0 and y >= 0)
                    assert(x < 100 and y < 100)
                except Exception:
                    x, y = 0, 0
                focusX.setText(str(x))
                focusY.setText(str(y))
                focusCb.setChecked(True)
                # print("set ({x},{y})".format(x=x, y=y))
        focusCb.toggled.connect(focusCbChanged)

        def focusChanged(text):
            focusCb.setChecked(False)
        focusX.textChanged.connect(focusChanged)
        focusY.textChanged.connect(focusChanged)

        focusD.setEnabled(False)

        unitSlider.setMaximum(10)
        unitSlider.setSingleStep(1)
        unitSlider.setTickInterval(1)
        unitSlider.setTickPosition(QSlider.TicksBelow)
        unitSlider.valueChanged.connect(
            lambda i: (unitLabel.setText("Unit: %-5s" % ("auto" if i == 0 else str(i))), self.sendCmd("AT+UNIT=%d\r" % (i))))
        unitSlider.setValue(0)

        fpsSlider.setMinimum(1)
        fpsSlider.setMaximum(30)
        fpsSlider.setSingleStep(1)
        fpsSlider.setTickInterval(1)
        fpsSlider.setTickPosition(QSlider.TicksBelow)
        fpsSlider.valueChanged.connect(
            lambda i: (fpsLabel.setText(" Fps: %-5d" % (i)), self.sendCmd("AT+FPS=%d\r" % (i))))
        fpsSlider.setValue(1)

        evSlider.setMinimum(0)
        evSlider.setMaximum(40000)
        evSlider.setSingleStep(2000)
        evSlider.setTickInterval(2000)
        evSlider.setTickPosition(QSlider.TicksBelow)
        evSlider.valueChanged.connect(
            lambda i: (evLabel.setText("  Ev: %-5s" % ("AE" if i == 0 else str(i))), self.sendCmd("AT+AE=%d\r" % (1 if i == 0 else 0)), self.sendCmd("AT+EV=%d\r" % (i))))
        evSlider.setValue(0)

    def remove(self):
        self.rmCallback(self)

    # def setRange(self, text):
    #     if text:
    #         self.config["xRange"] = float(text)

    # def setEnableRange(self, en):
    #     self.config["xRangeEnable"] = en

    def headerChanged(self, text):
        if self.config["header"] != text:
            self.headerBtn.setText(_("Set") + " *")
        else:
            self.headerBtn.setText(_("Set"))

    def setHeader(self, text):
        if text:
            try:
                textBytes = utils.str_to_bytes(
                    text, escape=True, encoding="utf-8")
                self.config["header"] = text
                self.headerBytes = textBytes
                self.headerBtn.setText(_("Set"))
            except Exception:
                self.hintSignal.emit("error", _("Error"), _("Format error"))

    def decodeData(self, data: bytes):
        '''
            @data bytes, protocol:
                         |    header(2B)    | len(2B) | command(1B) | output_mode(1B)     | sensor temp(1B) | driver temp(1B) |
                         |     0xCC  0xA0      |  >=18   |             | 0:Depth only, 1:+IR |    uint8        | uint8           |
                         -------------------------------------------------------------------------------------------------------
                         |    exposure time(4B)    | errcode(1B) | reserved1(1B) | res_rows(1B) | res_cols(1B) |
                         |[23:20][19:12][11:8][7:0]|             |       0x0     |     uint8    |     uint8    |
                         ---------------------------------------------------------------------------------------
                         |    Frame ID(2B)    | ISP ver(1B) | reserved3(1B) | frame data((len-16)B) | checksum(1B) | tail(1B) |
                         |     [11:0]         |     0x23    |       0x0     | xxxxxxxxxxxxxxxxxxxxx | sum of above |   0xDD   |

            @return haveFrame, dict {
                "frameID": {
                    "res": tunple(w, h)
                    "frameData": []
                }
            }
        '''
        # append data
        self.rawData += data
        # find header
        header = self.headerBytes
        idx = self.rawData.find(header)
        if idx < 0:
            return False, self.data
        self.rawData = self.rawData[idx:]
        # print(self.rawData)
        # check data length 2Byte
        dataLen = unpack("H", self.rawData[2: 4])[0]
        # print("len: "+str(dataLen))
        frameLen = len(header) + 2 + dataLen + 2
        frameDataLen = dataLen - 16

        if len(self.rawData) < frameLen:
            return False, self.data
        # get data
        frame = self.rawData[:frameLen]
        # print(frame.hex())
        self.rawData = self.rawData[frameLen:]

        frameTail = frame[-1]
        # print("tail: "+str(hex(frameTail)))
        _sum = frame[-2]
        # print("checksum: "+str(hex(_sum)))
        # check sum
        # spi has no checksum but i add one
        if frameTail != 0xdd and _sum != sum(frame[:frameLen - 2]) % 256:
            return False, self.data

        frameID = unpack("H", frame[16:18])[0]
        # print("frame ID: "+str(frameID))

        resR = unpack("B", frame[14:15])[0]
        resC = unpack("B", frame[15:16])[0]
        res = (resR, resC)
        # print(res)
        # frameData=[ unpack("H", frame[20+i:22+i])[0] for i in range(0, frameDataLen, 2) ]
        frameData = [unpack("B", frame[20+i:21+i])[0]
                     for i in range(0, frameDataLen, 1)]

        # if not frameID in self.data:
        #     self.data[frameID] = {
        #         "res": res,
        #         "frameData": frameData
        #     }
        # else:
        #     self.data[frameID]["res"] = res
        #     self.data[frameID]["frameData"] = frameData

        queueData = {
            'frameID': frameID,
            "res": res,
            "frameData": frameData
        }
        self.queue.put(queueData)

        return True, self.data

    def pickColor(self, name: str):
        if name in self.colors:
            return self.colors[name]
        else:
            if not self.notUsedColors:
                self.notUsedColors = self.builtinColors.copy()
            color = self.notUsedColors.pop(0)
            self.colors[name] = color
            return color

    def update(self, data: dict):
        # for k, v in data.items():
        if not self.queue.empty():
            v = self.queue.get()
            k = v['frameID']
            # print("draw frame:", k)
            res = v['res']
            frameData = v['frameData']
            arr = np.array(frameData, np.uint8)
            # arr[np.where(arr>3000)] = 0
            arr = arr.reshape(res)
            # arr = 255 - arr
            if self.onceShot:
                # print(rgba_img.shape)
                # DownloadsPATH = os.path.join(
                #     os.path.expanduser("~"), 'Downloads'
                DownloadsPATH = self.capTargetBtn.text()
                fname = "%d-unit%d" % ((time.time_ns() //
                                       1_000_000), self.unitSlider.value())
                Image.fromarray(arr).save(os.path.join(DownloadsPATH,
                                                       fname + ".bmp"))
                # plt.imsave(os.path.join(DownloadsPATH,
                #                         fname + ".bmp"), arr, cmap=plt.get_cmap('gray'))
                self.onceShot = False

            # cmap = plt.get_cmap('jet')
            # z = arr.transpose()
            # z = np.expand_dims(z, 2)

            # rgba_img = cmap(z)
            # rgba_img = np.squeeze(rgba_img)

            # print(rgba_img.shape)
            # print(v)

            if len(self.frames) != 0 and self.frames[len(self.frames)-1].shape != arr.T.shape:
                self.frames.clear()
            self.frames.append(arr)
            np_arrs = np.array(self.frames)

            if self.focus[0].isChecked():
                def compute_real_distance(val, unit):
                    ret = 0
                    if unit != 0:
                        ret = val * unit
                    else:
                        ret = int(val) / 5.1
                        ret *= ret

                    if len(self.distances) == 0:
                        for _ in range(8):
                            self.distances.append(ret)
                    else:
                        self.distances.append(ret)
                        self.distances = self.distances[-8:]
                    return int(np.mean(self.distances))

                self.focus[2].setText(str(compute_real_distance(np_arrs[len(
                    self.frames)-1][int(self.focus[1][0].text()), int(self.focus[1][1].text())], self.unitSlider.value()))+"mm")
                # print(self.focus[2].text())

                def draw_reticle_and_info(buf, x, y):
                    ll = buf.shape[0] // 25
                    u, d, l, r = min(ll*25, max(0, y-ll)), max(0, min(ll*25, y+ll+1)
                                                               ), min(ll*25, max(0, x-ll)), max(0, min(ll*25, x+ll+1))
                    buf[u:d, x] = 0x00
                    buf[y, l:r] = 0x00
                draw_reticle_and_info(np_arrs[len(self.frames)-1],
                                      int(self.focus[1][0].text()), int(self.focus[1][1].text()))

            self.imv.clear()
            self.imv.setImage(np_arrs)  # , autoRange=False)
            self.imv.setCurrentIndex(len(self.frames)-1)
            # color = self.pickColor(k)
            # self.colors[k] = color
            # self.curves[k] = self.p.plot(pen=pg.mkPen(color=color, width=2),
            #                             name=k,)
            # self.curves[k].setData(x=v["x"], y=v["y"])

            # imgItem = self.imv.getImageItem()
            # imgView = self.imv.getView()
            # print(imgItem.boundingRect())
            # print(imgView.viewGeometry())

    def capture(self):
        self.onceShot = True

    def onData(self, data: bytes):
        while True:
            haveFrame, allData = self.decodeData(data)
            if not haveFrame:
                break
            data = b''
        self.updateSignal.emit(allData)

    def sendCmd(self, cmd):
        send_bytes = utils.str_to_bytes(cmd, escape=True, encoding="ASCII")
        print(send_bytes)
        print(len(send_bytes))
        print()
        self.send(send_bytes)

